#include <stdio.h>
#include <sys/types.h>
#include <termios.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include "NNOCommandDefs.h"
#include "API.h"

#define MAX_BYTES_TO_READ	50
#define SERIAL_DEVICE	"/dev/ttys003"
#define MAX_PACKET_SIZE 64
#define MIN_PACKET_SIZE 64
#define PASS 0
#define FAIL -1
#define SUCCESS 1

static unsigned char OutputBuffer[MAX_PACKET_SIZE+2];
static unsigned char InputBuffer[sizeof(nnoMessageStruct)+MIN_PACKET_SIZE];
static unsigned char seqNum=0;
static uint8 UARTWriteBuffer[1024 * 8];
static uint8 ReadBuffer[1024 * 8];
#define UARTMAXSIZE 8192
static bool g_UARTConnected = true;
int UART_Supported[62] = {1,1,1,1,1,1,1,1,1,1,0,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,
                          1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1};

int uartMesglength = 0;
int UART_Device = 1;

int NNO_SetUARTConnected(bool connected)
{
    g_UARTConnected = connected;
    return PASS;
}

static int NNO_PrepReadCmd(uint32 cmd_key)
/**
 * This function is private to this file. Prepares the read-control command packet for the given command code and copies it to OutputBuffer.
 *
 * @param   cmd  - I - USB command code.
 *
 * @return  0 = PASS
 *          -1 = FAIL
 *
 */
{
    nnoMessageStruct msg;

    msg.head.flags.rw = 1; //Read
    msg.head.flags.reply = 1; //Host wants a reply from device
    msg.head.flags.dest = 0; //Projector Control Endpoint
    msg.head.flags.reserved = 0;
    msg.head.flags.resp = 0;
    msg.head.seq = 0;

    msg.payload.cmd = CMD_GET_CMD(cmd_key);
    msg.head.length = 2;

   if(cmd_key == NNO_CMD_FLASH_GET_CHKSUM)
    {
        msg.payload.data[2] = 0x00;
        msg.head.length += 1;
    }

    OutputBuffer[0]=0; // First byte is the report number
    memcpy(&OutputBuffer[1], &msg, (sizeof(msg.head)+sizeof(msg.payload.cmd) + msg.head.length));

    printf("SSUCCESS in NNO_PrepReadCmd\n");
    return PASS;
}

int Serial_Write(void *Data, unsigned long Length)
{
	printf("UART_Device = %d\n",UART_Device);
	if(UART_Device < 0){
		//THROW(ERR_NOT_INITIALIZED);
		printf("check line\n");
		printf("DEVICE NOT INITIALIZED _ Serial_Write\n");
		exit(1);
	}

	if(write(UART_Device, Data, Length) < 0){
		//THROW(FAIL);
		printf("Failed to write to device");
		exit(1);
	}
    printf("Serial_Write SUCCESS!");
    return SUCCESS;
}

int Serial_WaitForData(uint32 Timeout)
{
	fd_set InputFD;
	struct timeval TimeoutTimer;
	struct timeval *Timer;

	FD_ZERO(&InputFD);
	FD_SET(UART_Device, &InputFD);

	if(Timeout)
	{
		TimeoutTimer.tv_usec = Timeout;
		TimeoutTimer.tv_sec = 0;
		Timer = &TimeoutTimer;
	}
	else
	{
		Timer = NULL;
	}

	if(select(UART_Device + 1, &InputFD, NULL, NULL, Timer) == 0)
		//return ERR_TIMEOUT;
		return FAIL;

    return SUCCESS;
}

int Serial_TimeoutReadData(void *Data, uint32 Length, uint32 InitTimeout, uint32 Timeout)
{
	int Bytes;

	if(UART_Device < 0){
		//THROW(ERR_NOT_INITIALIZED);
		printf("check linke_ timeoutreaddata");
		perror("DEVICE NOT INITIALIZED _ Timeoutreaddata");
		exit(1);
	}
	while(1)
	{
		if(ioctl(UART_Device, FIONREAD, &Bytes) < 0){
			//THROW(FAIL);
			perror("FIONREAD ioctl failed\n");
			exit(6);
		}

		if(Bytes)
		{
			if(Bytes > Length)
				Bytes = Length;

			if(read(UART_Device, Data, Bytes) < 0){
				//THROW(FAIL);
				perror("Failed at read Uart_Device data");
				exit(6);
			}

			Length -= Bytes;
			Data = (uint8*)Data + Bytes;

			if(Length == 0)
				break;

			InitTimeout = Timeout;
		}

		if(Serial_WaitForData(InitTimeout)<0){
			//THROW_S(ERR_TIMEOUT);
			perror("ERR_TIMEOUT");
			exit(1);
		}

	}

    return SUCCESS;
}

int Serial_Read(void *Data, uint32 Length)
{
	return Serial_TimeoutReadData(Data, Length, 0, 0);
}



static int NNO_Read(int isUART)
/**
 * This function is private to this file. This function is called to write the read control command and then read back a complete HID packet over USB
 * to InputBuffer.
 *
 * @return  number of bytes read
 *          -2 = nack from target
 *          -1 = error reading
 *
 */
{

    nnoMessageStruct *pMsg = (nnoMessageStruct *)InputBuffer;
    short msg_len=MIN_PACKET_SIZE;
    printf("g_UARTConnected = %d\n",g_UARTConnected);
    if(g_UARTConnected && isUART)//UART_Supported[pMsg->payload.cmd])
    {
        memcpy(&UARTWriteBuffer[8],&OutputBuffer[1],65);
        pMsg  = (nnoMessageStruct *)&OutputBuffer[1];
        unsigned int Checksum = 0;
        for(int i = 8; i < pMsg->head.length + 12  ; i++)
        {
            Checksum += UARTWriteBuffer[i];
        }

        UARTWriteBuffer[4] = Checksum;
        UARTWriteBuffer[5] = Checksum >> 8;
        UARTWriteBuffer[6] = Checksum >> 16;
        UARTWriteBuffer[7] = Checksum >> 24;

        //write the head and tail
        UARTWriteBuffer[0] = UART_START_IND_BYTE_0;
        UARTWriteBuffer[1] = UART_START_IND_BYTE_1;
        UARTWriteBuffer[2] = UART_START_IND_BYTE_2;
        UARTWriteBuffer[3] = UART_START_IND_BYTE_3;

        UARTWriteBuffer[pMsg->head.length + 12] = UART_END_IND_BYTE_0;
        UARTWriteBuffer[pMsg->head.length + 13] = UART_END_IND_BYTE_1;
        UARTWriteBuffer[pMsg->head.length + 14] = UART_END_IND_BYTE_2;
        UARTWriteBuffer[pMsg->head.length + 15] = UART_END_IND_BYTE_3;


        nnouartMessageStruct *uMsg = (nnouartMessageStruct *)ReadBuffer;
        int read_length = 0;
        int ret;
        if(Serial_Write(UARTWriteBuffer,pMsg->head.length + 16) == 1)
        {
            printf("NNO_Read Serial_Write returns 1;");
            if(Serial_Read(ReadBuffer, 8 + sizeof(pMsg->head)) == 1)
            {
                read_length = uMsg->head.length;
                printf("NNO_Read Serial_Read returns 1;");
            }

            else
		perror("FAIL at Serial_Read!=1\n");
                return FAIL;

            if((ret = Serial_Read(&ReadBuffer[12], read_length + 4)) == 1)
            {
                //Serial.println("NNO_Read Serial_Read read_length+4 returns 1;");
                //skip the command name when copying from UART to USB structure
                memcpy(&InputBuffer[0],&ReadBuffer[8],4);
                memcpy(&InputBuffer[4],&ReadBuffer[14],read_length);
                //Serial.println("read_length");
                //Serial.println(read_length);
                return read_length;
            }
            else if(ret < 1)
            {
               return read_length;
            }
            else
                return FAIL;
        }
        return FALSE;
    }
//    int ret_val;
//    int bytes_read=0;
//
//    if(USB_Write((void *)OutputBuffer) > 0)
//    {
//        while(bytes_read < msg_len)
//        {
//            ret_val =  USB_Read(&InputBuffer[bytes_read]);
//            if(ret_val < 0)
//            {
//                return bytes_read;
//            }
//            else if (ret_val == 0)
//                return NNO_READ_TIMEOUT;
//
//            if(bytes_read == 0)//On reading the header of a message, determine the message length or find if a NACK was received
//            {
//                if(pMsg->head.flags.resp == NNO_RESP_BUSY)
//                    return NNO_CMD_BUSY;
//                else if((pMsg->head.flags.resp == NNO_RESP_ERROR) || (pMsg->head.length == 0))
//                    return NNO_CMD_NACK;
//
//                msg_len = sizeof(pMsg->head) + pMsg->head.length;
//            }
//            bytes_read += ret_val;
//        }
//        return bytes_read;
//    }
    return FAIL;

}

int NNO_TestEEPROM(void)
/**
 * Sends command to perform EEPROM test.
 *
 * (USB CMD OPCODE: BYTE1: 0x01, BYTE2: 0x01)
 *
 * @ingroup testAPI
 *
 * @return  PASS or FAIL
 *
 */
{
    nnoMessageStruct msg;
    int ret_val;
    printf("In NNO_TESTEEPROM\n");
    NNO_PrepReadCmd(NNO_CMD_EEPROM_TEST);
    if((ret_val = NNO_Read(1)) > 0)
    {
	printf("ret_val>0, is: %d",ret_val);
        memcpy(&msg, InputBuffer, 65);
        if(msg.payload.data[0] == 0)
            return PASS;
        else
            return FAIL;
    }
    printf("ret_val = %d",ret_val);
    //Serial.println("return ret_val");
    return ret_val;
}

int main()
{
	struct termios serial_port_settings;
	int fd;
	int retval;
	char buf[256];
	int i;

	fd = open(SERIAL_DEVICE, O_RDWR);
	if (fd < 0) {
		perror("Failed to open SERIAL_DEVICE");
		exit(1);
	}
	else{
		printf("open serial at: 003\n");

	}
	//UART_Device=fd+1;
	retval = tcgetattr(fd, &serial_port_settings);
	if (retval < 0) {
		perror("Failed to get termios structure");
		exit(2);
	}
	else{
		printf("tcgetattr at 003\n");
	}

	//setting baud rate to B115200
	retval = cfsetospeed(&serial_port_settings, B115200);
	if (retval < 0) {
		perror("Failed to set 115200 output baud rate");
		exit(3);
	}
	retval = cfsetispeed(&serial_port_settings, B115200);
	if (retval < 0) {
		perror("Failed to set 115200 input baud rate");
		exit(4);
	}
	serial_port_settings.c_lflag &= ~(ICANON);
	serial_port_settings.c_lflag &= ~(ECHO | ECHOE);
	retval = tcsetattr(fd, TCSANOW, &serial_port_settings);
	if (retval < 0) {
		perror("Failed to set serial attributes");
		exit(5);
	}
	printf("Successfully set the baud rate\n");
	NNO_SetUARTConnected(true);
	int nnotestres=NNO_TestEEPROM();	
        printf("NNO_TestEEPROM() = %d\n",nnotestres);
	return 0;

}
